/*
    Fichier voirmail.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence.fr

    Traductions des droits d'utilisation dans les fichiers :
    licence.de , licence.en , licence.es , licence.it
    licence.nl , licence.pt , licence.eo , licence.eo-utf


    Ce programme affiche un mail en attente de tlchargement
    (ou de destruction).
    Un fichier de configuration est utilis pour se connecter  la
    boite aux lettres.
    Le mail  afficher est repr par son numro d'ordre dans la
    liste des mails sur le serveur.
*/

#define appli   // pour la dclaration de variables globales  l'application

#include <stdio.h>
#include <string.h>
#include "buflect.h"
#include "pop.h"
#include "encodage.h"
#include "trtentete.h"
#include "trtligne.h"
#include "trtbordure.h"
#include "trtsection.h"


/* prototypes */
void li_entete (int numes);
void aff_mail ();
void aff_texte ();
int  fin_mail ();


/* variables globales  l'application utilises par modepage
   (pour viter des tonnes de passages de paramtres) */

int ctypeorig; // type principal du mail
int ctype;     // type de la section courante
int opth;      // affichage de la section text/html au lieu de text/plain


/* programme principal */

int main (int nbarg, char *varg[])
{
    char ficonf [80];   // fichier de configuration du systme de mail
    FILE *fconf;        // descripteur du fichier de configuration
    int  numes;         // numro du mail courant

    // si commande lance avec l'option -h
    if (--nbarg > 1 && strcmp (varg [1], "-h") == 0)
        // mmoriser cette option
        opth = 1;
    else
        // sinon pas d'option -h
        opth = 0;

    // controle du nombre d'arguments
    if (nbarg == opth + 2)
    {
        // rcuprer le nom du fichier de configuration
        strcpy (ficonf, varg [nbarg]);

        if (ficonf [strlen (ficonf) - 4] != '.')
            strcat (ficonf, ".cfg");

        // ouvrir le fichier de configuration
        fconf = fopen (ficonf, "r");

        if (fconf)
        {
            // connexion sur le compte mail du serveur pop
            if (connect_pop (fconf))
            {
                // rcuprer le numro de mail
                numes = atoi (varg [opth + 1]);

                // afficher ce mail si possible
                if (numes > 0 && numes <= nbmails ())
                    li_entete (numes);
                else if (numes > 0)
                    fputs ("Numro de mail trop grand\n", stderr);
                else
                    fprintf (stderr, "Numro de mail %s incorrect\n",
                                                     varg [opth + 1]);

                // se dconnecter proprement du serveur pop
                deconnect_pop ();
            }

            // on n'a plus besoin du fichier de configuration
            fclose (fconf);
        }
        else
            fprintf (stderr, "Fichier %s non trouv\n", ficonf);
    }
    else
        fprintf (stderr,
         "Syntaxe : %s [-h] numro_message fichier_configuration\n", varg [0]);

    // pour faire plaisir  gcc qui veut une fonction main de type int
    return (0);
}


/* lit le message choisi et l'affiche */

void li_entete (int numes)
{
    char bufw [80]; // buffer d'envoi d'une requte
    // buffers pour affichage ordonn des caractristiques du message
    char bufFrom [80], bufTo [80], bufCc [80], bufReply [80], bufSubject [80];
    int  analsuiv;   // indicateur : il faudra traiter la ligne suivante


    // initialisation
    bufFrom [0]    = '\0';
    bufTo [0]      = '\0';
    bufCc [0]      = '\0';
    bufReply [0]   = '\0';
    bufSubject [0] = '\0';
    analsuiv       =   0;
    nbordures      =   0;
    ctypeorig = TextPlain;  // valeur par dfaut

    // demande de lecture du message
    sprintf (bufw, "RETR %d", numes);
    env_pop (bufw);

    printf ("Message %d\n", numes);

    // lecture et mmorisation des caractristiques du message
    do
    {
        // lire une ligne de l'entte du message
        lire_pop ();

        // rcupration si ncessaire du charset ou de la
        // bordure dans la 2me ligne du champ Content-Type
        if (analsuiv)
        {
            // si mode multipart
            if (ctypeorig & Multipart)
                // mmoriser la bordure
                mem_boundary ();
            // sinon
            else
                // mmoriser le jeu de caractres
                mem_charset ();

            // l'analyse ne se poursuivra pas sur les lignes suivantes
            analsuiv = 0;
        }

        // mmorisation des caractristiques importantes
        if (start ("Date"))
            puts (buf_lect);   // la date est affiche directement
        else if (start ("From"))
            memconvbuf (bufFrom);
        else if (start ("To"))
            memconvbuf (bufTo);
        else if (start ("Cc"))
            memconvbuf (bufCc);
        else if (start ("Reply-To"))
            memconvbuf (bufReply);
        else if (start ("Subject"))
            memconvbuf (bufSubject);
        else if (start ("Content-Type"))
        {
            // rcuprer la valeur du champ Content-Type
            ctypeorig = recup_ctype ();

            // si ce type correspond au texte du mail
            if (ctypeorig == TextPlain)
            {
                // mmoriser le jeu de caractres
                if (! mem_charset ())
                    //  partir de cette ligne ou de la suivante
                    analsuiv = 1;
            }
            // sinon si mode multipart
            else if (ctypeorig & Multipart)
            {
                // mmoriser la bordure
                if (! mem_boundary ())
                    //  partir de cette ligne ou de la suivante
                    analsuiv = 1;
            }
        }
        else if (start ("Content-Transfer-Encoding"))
            mem_encodage ();
    }
    while (buf_lect [0] != '\0');  // lecture entte termine si ligne vide

    // affichage de certaines caractristiques mmorises
    if (bufFrom [0])
        puts (bufFrom);
    if (bufTo [0])
        puts (bufTo);
    if (bufCc [0])
        puts (bufCc);
    if (bufReply [0])
        puts (bufReply);
    if (bufSubject [0])
        puts (bufSubject);

    // sauter la ligne blanche qui prcde le message
    lire_pop ();

    // affichage du contenu du mail
    if (ctypeorig & Multipart)
        // il faudra analyser les sections
        aff_mail ();
    else
        // un affichage simple suffit
        aff_texte ();
}


/* affichage du contenu du mail */

void aff_mail ()
{
    // pour l'instant, le type principal du mail est le type courant
    ctype = ctypeorig;

    // si mode multipart/mixed, mmoriser le type et
    // la bordure ventuelle de la section suivante
    if (ctypeorig == MultipMixed)
        mem_soustype ();

    // se positionner sur la section texte (ou text/html si option -h)
    // du mail (si l'on n'y est pas)
    if (opth)
        posit_texthtm ();
    else
        posit_texte ();

    // message d'avertissement ventuel
    if (encodage_texte == Base64)
        puts ("\n********* Texte non standard : encodage Base64 *********");

    // gnrer une ligne de sparation avec l'entte
    putchar ('\n');

    // si structure du mail non conforme, message d'erreur
    if (fin_mail ())
    {
        if (opth)
            puts ("Pas de zone texte html dans ce mail !!!");
        else
            puts ("Pas de zone texte dans ce mail !!!");
    }

    // lecture et affichage du corps du message
    // on s'arrte en fin de mail, ou sur la prochaine bordure
    while (! fin_mail () && (nbordures == 0 || ! surbordure ()))
    {
        // mise en forme et affichage de la dernire ligne lue
        majligne ();
        printf ("%s", buf_lect);

        // et lecture de la suivante
        lire_pop ();

        // si mode multipart/report, on saute les enttes de section
        if (ctypeorig == MultipRep && surbordure ())
        {
            do
                lire_pop ();
            while (*buf_lect && ! fin_mail ());
        }
    }

    // si mode multipart mixed on va lister les pices jointes
    if (ctypeorig == MultipMixed)
        liste_pj ();
}



/* lecture et affichage d'un mail de type text/plain */

void aff_texte ()
{
    // message d'avertissement ventuel
    if (encodage_texte == Base64)
        puts ("\n********* Texte non standard : encodage Base64 *********");

    // rpter
    do
    {
        // mise en forme et affichage de la dernire ligne lue
        majligne ();
        printf ("%s", buf_lect);

        // et lecture de la suivante
        lire_pop ();
    }
    // on s'arrte en fin de mail
    while (! fin_mail ());
}



// indique si le mail a t lu entirement

int fin_mail ()
{
    return (buf_lect [0] == '.' && buf_lect [1] == '\0');
}



/* Ce qui suit est une adaptation des fonctions de trtsection
   en remplaant la fonction lire_fmail par lire_pop */


/* avancer jusqu' la section suivante si
   on n'est pas sur un marqueur de section */

void prochaine_section ()
{
    // tant qu'on n'est pas sur une bordure
    while (! surbordure ())
    {
        // avancer d'une ligne
        lire_pop ();

        // si problme de structure du mail (fin de mail rencontre)
        if (fin_mail ())
            return;   // sortir
    }

    // si bordure trouve, lire la ligne suivante
    if (buf_lect [0] == '-')
        lire_pop ();
}



/* mmorise le sous type de la premire section imbriqu
   dans un mail multipart/mixed et sa bordure s'il y a lieu */

void mem_soustype ()
{
    // avancer jusqu' la prochaine section
    prochaine_section ();

    // termin si problme de structure du mail
    if (fin_mail ())
        return;

    // mmoriser les caractristiques de la section
    do
    {
        // content-type
        if (start ("Content-Type"))
        {
            // mmoriser le type de la section
            ctype = recup_ctype ();

            // si on a trouv un autre mode multipart imbriqu
            if (ctype & Multipart)
            {
                // mmoriser sa bordure
                while (*buf_lect && ! mem_boundary ())
                    lire_pop ();
            }
        }

        // mode d'encodage
        else if (start ("Content-Transfer-Encoding"))
            mem_encodage ();

        // passer  la ligne suivante
        lire_pop ();
    }
    // on s'arrtera lorsqu'on aura trouv un nouveau type
    // ou en fin d'entte de section, (ou de mail)
    while (ctype == MultipMixed && buf_lect [0] != '\0');
}



// se positionner sur la section texte du mail (si l'on n'y est pas)

void posit_texte ()
{
    posit_section (TextPlain);
}



// se positionner sur la section text/html du mail (si l'on n'y est pas)

void posit_texthtm ()
{
    posit_section (TextHtml);
}



// se positionner (si l'on n'y est pas) sur la section text/plain
// ou text/html du mail, en fonction du paramtre  typetexte

void posit_section (int typetexte)
{
    int OK_charset = 0;  // indicateur : charset trouv


    // sortie directe si problme de structure du mail (fin de mail)
    if (fin_mail ())
        return;

    // tant qu'on n'a pas trouv la section contenant le texte du mail
    while (ctype != typetexte)
    {
        // si une bordure a t mmorise
        if (nbordures)
            // se positionner sur la section suivante
            prochaine_section ();

        // sortie directe si section non trouve (pb de structure du mail)
        if (fin_mail ())
            return;

        // mmoriser ses caractristiques
        do
        {
            // content-type et charset ventuel
            if (start ("Content-Type"))
            {
                ctype = recup_ctype ();
                OK_charset = mem_charset ();
            }

            // mode d'encodage
            else if (start ("Content-Transfer-Encoding"))
                mem_encodage ();

            // avancer d'une ligne
            lire_pop ();
        }
        // jusqu' type trouv ou fin d'entte
        while (ctype != typetexte && buf_lect [0] != '\0' && !fin_mail ());
    }

    // on est sur la section contenant le texte du mail
    // on va mmoriser le charset et le mode d'encodage s'il y a lieu
    while (buf_lect [0] != '\0')
    {
        if (start ("Content-Transfer-Encoding"))
            mem_encodage ();

        else if (! OK_charset)
            OK_charset = mem_charset ();

        // passer  la ligne suivante
        lire_pop ();
    }
}



// lister les pices jointes

void liste_pj ()
{
    int  posjoint;        // position du nom du fichier joint
    char nomjoint [80];   // nom du fichier joint
    int  nbext;           // nombre d'extentions du fichier
    int  i;               // compteur


    // on ne conserve que la bordure de premier niveau
    // (permet de sauter le texte HTML en mode multipart/alternative)
    nbordures = 1;

    // tant que non fin de mail
    while (! fin_mail ())
    {
        // se positionner sur la prochaine section si on n'y est pas dj
        prochaine_section ();

        // termin s'il n'y en a plus
        if (fin_mail ())
           return;

        // chercher et mmoriser le nom du fichier joint s'il y en a un
        do
        {
            majligne ();
            posjoint = posnomfic ();

            // si trouv
            if (posjoint)
            {
                // rcuprer son nom
                strcpy (nomjoint, "-> Fichier joint : ");
                i = strlen (nomjoint);
                nbext = 0;

                // recopie du nom de fichier
                while (buf_lect [posjoint] != '"' && buf_lect [posjoint])
                {
                    // sans les (horribles) blancs qu'il pourrait contenir !!!
                    if (buf_lect [posjoint] != ' ')
                        nomjoint [i++] = buf_lect [posjoint];

                    // passer au caractre suivant
                    posjoint++;

                    // on compte les . (nombre d'extentions) du fichier
                    //  partir du 2me caractre du nom
                    if (buf_lect [posjoint] == '.')
                        nbext ++;
                }

                // terminer la chaine gnre
                nomjoint [i] = '\0';
            }

            // passer  la ligne suivante
            lire_pop ();
        }
        while (! posjoint && buf_lect [0] != '\0' && ! fin_mail ());

        // si fichier joint trouv
        if (posjoint)
        {
            // et que son nom comporte plusieurs extentions
            // if (nbext > 1)  *** test devenu insuffisant en 2004 ***

            // actuellement tous les fichiers joints avec extention peuvent
            // contenir des virus, on accepte quand mme les .exe
            if (nbext > 1 || (nbext == 1 && strcmp (nomjoint + i - 4, ".exe\n")
                             != 0 && strcmp (nomjoint + i - 4, ".EXE\n") != 0))
            {
                // se positionner sur la premire ligne du contenu du fichier
                while (buf_lect [0] != '\0' && ! fin_mail ())
                    lire_pop ();

                if (! fin_mail ())
                    lire_pop ();

                // si le fichier commence par les caractres MZ
                // (on teste l'encodage base64 correspondant)
                if (buf_lect [0] == 'T' && buf_lect [1] == 'V'
                 && buf_lect [2] >= 'o' && buf_lect [2] <= 'r')
                    // il peut s'agir d'un virus MS-DOS / Windows
                    strcpy (nomjoint + i, "     *** VIRUS ? ***");
            }

            // afficher le nom du fichier joint trouv
            puts (nomjoint);
        }
    }
}
